@@ -1,43 +1,75 @@
 # Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
-# For details: https://bitbucket.org/ned/coveragepy/src/default/NOTICE.txt
+# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt
 
 """Base test case class for coverage.py testing."""
 
 import contextlib
 import datetime
+import functools
 import glob
 import os
+import os.path
 import random
 import re
 import shlex
-import shutil
 import sys
+import types
 
+import pytest
 from unittest_mixins import (
     EnvironmentAwareMixin, StdStreamCapturingMixin, TempDirMixin,
     DelayedAssertionMixin,
 )
 
 import coverage
-from coverage.backunittest import TestCase
+from coverage import env
+from coverage.backunittest import TestCase, unittest
 from coverage.backward import StringIO, import_local_file, string_class, shlex_quote
-from coverage.backward import invalidate_import_caches
 from coverage.cmdline import CoverageScript
-from coverage.debug import _TEST_NAME_FILE, DebugControl
+from coverage.misc import StopEverything
 
-from tests.helpers import run_command
+from tests.helpers import arcs_to_arcz_repr, arcz_to_arcs
+from tests.helpers import run_command, SuperModuleCleaner
 
 
 # Status returns for the command line.
 OK, ERR = 0, 1
 
+# The coverage/tests directory, for all sorts of finding test helping things.
+TESTS_DIR = os.path.dirname(__file__)
+
+
+def convert_skip_exceptions(method):
+    """A decorator for test methods to convert StopEverything to SkipTest."""
+    @functools.wraps(method)
+    def _wrapper(*args, **kwargs):
+        try:
+            result = method(*args, **kwargs)
+        except StopEverything:
+            raise unittest.SkipTest("StopEverything!")
+        return result
+    return _wrapper
+
+
+class SkipConvertingMetaclass(type):
+    """Decorate all test methods to convert StopEverything to SkipTest."""
+    def __new__(cls, name, bases, attrs):
+        for attr_name, attr_value in attrs.items():
+            if attr_name.startswith('test_') and isinstance(attr_value, types.FunctionType):
+                attrs[attr_name] = convert_skip_exceptions(attr_value)
+
+        return super(SkipConvertingMetaclass, cls).__new__(cls, name, bases, attrs)
+
+
+CoverageTestMethodsMixin = SkipConvertingMetaclass('CoverageTestMethodsMixin', (), {})
 
 class CoverageTest(
     EnvironmentAwareMixin,
     StdStreamCapturingMixin,
     TempDirMixin,
     DelayedAssertionMixin,
-    TestCase
+    CoverageTestMethodsMixin,
+    TestCase,
 ):
     """A base class for coverage.py test cases."""
 
@@ -47,19 +79,31 @@ class CoverageTest(
     # Tell newer unittest implementations to print long helpful messages.
     longMessage = True
 
+    # Let stderr go to stderr, pytest will capture it for us.
+    show_stderr = True
+
+    # Temp dirs go to $TMPDIR/coverage_test/*
+    temp_dir_prefix = "coverage_test/"
+    if os.getenv('COVERAGE_ENV_ID'):
+        temp_dir_prefix += "{}/".format(os.getenv('COVERAGE_ENV_ID'))
+
+    # Keep the temp directories if the env says to.
+    # $set_env.py: COVERAGE_KEEP_TMP - Keep the temp directories made by tests.
+    keep_temp_dir = bool(int(os.getenv("COVERAGE_KEEP_TMP", "0")))
+
     def setUp(self):
         super(CoverageTest, self).setUp()
 
+        self.module_cleaner = SuperModuleCleaner()
+
         # Attributes for getting info about what happened.
         self.last_command_status = None
         self.last_command_output = None
         self.last_module_name = None
 
-        if _TEST_NAME_FILE:                                 # pragma: debugging
-            with open(_TEST_NAME_FILE, "w") as f:
-                f.write("%s_%s" % (
-                    self.__class__.__name__, self._testMethodName,
-                ))
+    def xfail(self, msg):
+        """Mark this test as an expected failure."""
+        pytest.xfail(msg)
 
     def clean_local_file_imports(self):
         """Clean up the results of calls to `import_local_file`.
@@ -68,27 +112,7 @@ class CoverageTest(
         one test.
 
         """
-        # So that we can re-import files, clean them out first.
-        self.cleanup_modules()
-        # Also have to clean out the .pyc file, since the timestamp
-        # resolution is only one second, a changed file might not be
-        # picked up.
-        for pyc in glob.glob('*.pyc'):
-            os.remove(pyc)
-        if os.path.exists("__pycache__"):
-            shutil.rmtree("__pycache__")
-
-        invalidate_import_caches()
-
-    def import_local_file(self, modname, modfile=None):
-        """Import a local file as a module.
-
-        Opens a file in the current directory named `modname`.py, imports it
-        as `modname`, and returns the module object. `modfile` is the file to
-        import if it isn't in the current directory.
-
-        """
-        return import_local_file(modname, modfile)
+        self.module_cleaner.clean_local_file_imports()
 
     def start_import_stop(self, cov, modname, modfile=None):
         """Start coverage, import a file, then stop coverage.
@@ -103,7 +127,7 @@ class CoverageTest(
         cov.start()
         try:                                    # pragma: nested
             # Import the Python file, executing it.
-            mod = self.import_local_file(modname, modfile)
+            mod = import_local_file(modname, modfile)
         finally:                                # pragma: nested
             # Stop coverage.py.
             cov.stop()
@@ -114,49 +138,11 @@ class CoverageTest(
         self.last_module_name = 'coverage_test_' + str(random.random())[2:]
         return self.last_module_name
 
-    # Map chars to numbers for arcz_to_arcs
-    _arcz_map = {'.': -1}
-    _arcz_map.update(dict((c, ord(c) - ord('0')) for c in '123456789'))
-    _arcz_map.update(dict(
-        (c, 10 + ord(c) - ord('A')) for c in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
-    ))
-
-    def arcz_to_arcs(self, arcz):
-        """Convert a compact textual representation of arcs to a list of pairs.
-
-        The text has space-separated pairs of letters.  Period is -1, 1-9 are
-        1-9, A-Z are 10 through 36.  The resulting list is sorted regardless of
-        the order of the input pairs.
-
-        ".1 12 2." --> [(-1,1), (1,2), (2,-1)]
-
-        Minus signs can be included in the pairs:
-
-        "-11, 12, 2-5" --> [(-1,1), (1,2), (2,-5)]
-
-        """
-        arcs = []
-        for pair in arcz.split():
-            asgn = bsgn = 1
-            if len(pair) == 2:
-                a, b = pair
-            else:
-                assert len(pair) == 3
-                if pair[0] == '-':
-                    _, a, b = pair
-                    asgn = -1
-                else:
-                    assert pair[1] == '-'
-                    a, _, b = pair
-                    bsgn = -1
-            arcs.append((asgn * self._arcz_map[a], bsgn * self._arcz_map[b]))
-        return sorted(arcs)
-
-    def assert_equal_args(self, a1, a2, msg=None):
+    def assert_equal_arcs(self, a1, a2, msg=None):
         """Assert that the arc lists `a1` and `a2` are equal."""
         # Make them into multi-line strings so we can see what's going wrong.
-        s1 = "\n".join(repr(a) for a in a1) + "\n"
-        s2 = "\n".join(repr(a) for a in a2) + "\n"
+        s1 = arcs_to_arcz_repr(a1)
+        s2 = arcs_to_arcz_repr(a2)
         self.assertMultiLineEqual(s1, s2, msg)
 
     def check_coverage(
@@ -190,11 +176,11 @@ class CoverageTest(
         self.make_file(modname + ".py", text)
 
         if arcs is None and arcz is not None:
-            arcs = self.arcz_to_arcs(arcz)
+            arcs = arcz_to_arcs(arcz)
         if arcs_missing is None:
-            arcs_missing = self.arcz_to_arcs(arcz_missing)
+            arcs_missing = arcz_to_arcs(arcz_missing)
         if arcs_unpredicted is None:
-            arcs_unpredicted = self.arcz_to_arcs(arcz_unpredicted)
+            arcs_unpredicted = arcz_to_arcs(arcz_unpredicted)
 
         # Start up coverage.py.
         cov = coverage.Coverage(branch=True)
@@ -238,19 +224,19 @@ class CoverageTest(
 
         if arcs is not None:
             with self.delayed_assertions():
-                self.assert_equal_args(
-                    analysis.arc_possibilities(), arcs,
-                    "Possible arcs differ: minus is actual, plus is expected"
+                self.assert_equal_arcs(
+                    arcs, analysis.arc_possibilities(),
+                    "Possible arcs differ: minus is expected, plus is actual"
                 )
 
-                self.assert_equal_args(
-                    analysis.arcs_missing(), arcs_missing,
-                    "Missing arcs differ: minus is actual, plus is expected"
+                self.assert_equal_arcs(
+                    arcs_missing, analysis.arcs_missing(),
+                    "Missing arcs differ: minus is expected, plus is actual"
                 )
 
-                self.assert_equal_args(
-                    analysis.arcs_unpredicted(), arcs_unpredicted,
-                    "Unpredicted arcs differ: minus is actual, plus is expected"
+                self.assert_equal_arcs(
+                    arcs_unpredicted, analysis.arcs_unpredicted(),
+                    "Unpredicted arcs differ: minus is expected, plus is actual"
                 )
 
         if report:
@@ -262,11 +248,28 @@ class CoverageTest(
         return cov
 
     @contextlib.contextmanager
-    def assert_warnings(self, cov, warnings):
-        """A context manager to check that particular warnings happened in `cov`."""
+    def assert_warnings(self, cov, warnings, not_warnings=()):
+        """A context manager to check that particular warnings happened in `cov`.
+
+        `cov` is a Coverage instance.  `warnings` is a list of regexes.  Every
+        regex must match a warning that was issued by `cov`.  It is OK for
+        extra warnings to be issued by `cov` that are not matched by any regex.
+        Warnings that are disabled are still considered issued by this function.
+
+        `not_warnings` is a list of regexes that must not appear in the
+        warnings.  This is only checked if there are some positive warnings to
+        test for in `warnings`.
+
+        If `warnings` is empty, then `cov` is not allowed to issue any
+        warnings.
+
+        """
         saved_warnings = []
-        def capture_warning(msg):
+        def capture_warning(msg, slug=None, once=False):        # pylint: disable=unused-argument
             """A fake implementation of Coverage._warn, to capture warnings."""
+            # NOTE: we don't implement `once`.
+            if slug:
+                msg = "%s (%s)" % (msg, slug)
             saved_warnings.append(msg)
 
         original_warn = cov._warn
@@ -274,15 +277,25 @@ class CoverageTest(
 
         try:
             yield
-        except:
+        except:                     # pylint: disable=try-except-raise
             raise
         else:
-            for warning_regex in warnings:
-                for saved in saved_warnings:
-                    if re.search(warning_regex, saved):
-                        break
-                else:
-                    self.fail("Didn't find warning %r in %r" % (warning_regex, saved_warnings))
+            if warnings:
+                for warning_regex in warnings:
+                    for saved in saved_warnings:
+                        if re.search(warning_regex, saved):
+                            break
+                    else:
+                        self.fail("Didn't find warning %r in %r" % (warning_regex, saved_warnings))
+                for warning_regex in not_warnings:
+                    for saved in saved_warnings:
+                        if re.search(warning_regex, saved):
+                            self.fail("Found warning %r in %r" % (warning_regex, saved_warnings))
+            else:
+                # No warnings expected. Raise if any warnings happened.
+                if saved_warnings:
+                    self.fail("Unexpected warnings: %r" % (saved_warnings,))
+        finally:
             cov._warn = original_warn
 
     def nice_file(self, *fparts):
@@ -306,6 +319,13 @@ class CoverageTest(
         msg = "File %r shouldn't exist" % fname
         self.assertTrue(not os.path.exists(fname), msg)
 
+    def assert_file_count(self, pattern, count):
+        """Assert that there are `count` files matching `pattern`."""
+        files = sorted(glob.glob(pattern))
+        msg = "There should be {} files matching {!r}, but there are these: {}"
+        msg = msg.format(count, pattern, files)
+        self.assertEqual(len(files), count, msg)
+
     def assert_starts_with(self, s, prefix, msg=None):
         """Assert that `s` starts with `prefix`."""
         if not s.startswith(prefix):
@@ -314,12 +334,10 @@ class CoverageTest(
     def assert_recent_datetime(self, dt, seconds=10, msg=None):
         """Assert that `dt` marks a time at most `seconds` seconds ago."""
         age = datetime.datetime.now() - dt
-        # Python2.6 doesn't have total_seconds :(
-        self.assertEqual(age.days, 0, msg)
-        self.assertGreaterEqual(age.seconds, 0, msg)
-        self.assertLessEqual(age.seconds, seconds, msg)
+        self.assertGreaterEqual(age.total_seconds(), 0, msg)
+        self.assertLessEqual(age.total_seconds(), seconds, msg)
 
-    def command_line(self, args, ret=OK, _covpkg=None):
+    def command_line(self, args, ret=OK):
         """Run `args` through the command line.
 
         Use this when you want to run the full coverage machinery, but in the
@@ -331,8 +349,7 @@ class CoverageTest(
         Returns None.
 
         """
-        script = CoverageScript(_covpkg=_covpkg)
-        ret_actual = script.command_line(shlex.split(args))
+        ret_actual = command_line(args)
         self.assertEqual(ret_actual, ret)
 
     coverage_command = "coverage"
@@ -367,7 +384,7 @@ class CoverageTest(
             Python interpreter.
 
         * "coverage" is replaced with the command name for the main
-            Coverage.py program.
+            coverage.py program.
 
         Returns a pair: the process' exit status and its stdout/stderr text,
         which are also stored as `self.last_command_status` and
@@ -376,7 +393,7 @@ class CoverageTest(
         """
         # Make sure "python" and "coverage" mean specifically what we want
         # them to mean.
-        split_commandline = cmd.split(" ", 1)
+        split_commandline = cmd.split()
         command_name = split_commandline[0]
         command_args = split_commandline[1:]
 
@@ -386,24 +403,40 @@ class CoverageTest(
             # get executed as "python3.3 foo.py". This is important because
             # Python 3.x doesn't install as "python", so you might get a Python
             # 2 executable instead if you don't use the executable's basename.
-            command_name = os.path.basename(sys.executable)
+            command_words = [os.path.basename(sys.executable)]
 
-        if command_name == "coverage":
-            # The invocation requests the Coverage.py program.  Substitute the
-            # actual Coverage.py main command name.
-            command_name = self.coverage_command
+        elif command_name == "coverage":
+            if env.JYTHON:                  # pragma: only jython
+                # Jython can't do reporting, so let's skip the test now.
+                if command_args and command_args[0] in ('report', 'html', 'xml', 'annotate'):
+                    self.skipTest("Can't run reporting commands in Jython")
+                # Jython can't run "coverage" as a command because the shebang
+                # refers to another shebang'd Python script. So run them as
+                # modules.
+                command_words = "jython -m coverage".split()
+            else:
+                # The invocation requests the coverage.py program.  Substitute the
+                # actual coverage.py main command name.
+                command_words = [self.coverage_command]
 
-        cmd = " ".join([shlex_quote(command_name)] + command_args)
+        else:
+            command_words = [command_name]
+
+        cmd = " ".join([shlex_quote(w) for w in command_words] + command_args)
 
         # Add our test modules directory to PYTHONPATH.  I'm sure there's too
         # much path munging here, but...
+        pythonpath_name = "PYTHONPATH"
+        if env.JYTHON:
+            pythonpath_name = "JYTHONPATH"          # pragma: only jython
+
         testmods = self.nice_file(self.working_root(), 'tests/modules')
         zipfile = self.nice_file(self.working_root(), 'tests/zipmods.zip')
-        pypath = os.getenv('PYTHONPATH', '')
+        pypath = os.getenv(pythonpath_name, '')
         if pypath:
             pypath += os.pathsep
         pypath += testmods + os.pathsep + zipfile
-        self.set_environ('PYTHONPATH', pypath)
+        self.set_environ(pythonpath_name, pypath)
 
         self.last_command_status, self.last_command_output = run_command(cmd)
         print(self.last_command_output)
@@ -438,12 +471,33 @@ class CoverageTest(
         """Return the last line of `report` with the spaces squeezed down."""
         return self.squeezed_lines(report)[-1]
 
+    def get_measured_filenames(self, coverage_data):
+        """Get paths to measured files.
 
-class DebugControlString(DebugControl):
-    """A `DebugControl` that writes to a StringIO, for testing."""
-    def __init__(self, options):
-        super(DebugControlString, self).__init__(options, StringIO())
+        Returns a dict of {filename: absolute path to file}
+        for given CoverageData.
+        """
+        return {os.path.basename(filename): filename
+                for filename in coverage_data.measured_files()}
 
-    def get_output(self):
-        """Get the output text from the `DebugControl`."""
-        return self.output.getvalue()
+
+class UsingModulesMixin(object):
+    """A mixin for importing modules from tests/modules and tests/moremodules."""
+
+    def setUp(self):
+        super(UsingModulesMixin, self).setUp()
+
+        # Parent class saves and restores sys.path, we can just modify it.
+        sys.path.append(self.nice_file(TESTS_DIR, 'modules'))
+        sys.path.append(self.nice_file(TESTS_DIR, 'moremodules'))
+
+
+def command_line(args):
+    """Run `args` through the CoverageScript command line.
+
+    Returns the return code from CoverageScript.command_line.
+
+    """
+    script = CoverageScript()
+    ret = script.command_line(shlex.split(args))
+    return ret
